/*
** This is the source code to a "wish" that has SimpleFS builtin
**
** The startup script is located as follows:
**
**   (1)  Open the executable 
**   (2)  DeObfuscate the startup Tcl script.
**   (3)  Do VFS_Init, Memchan_Init and maybe Dde and Registry_Init
**   (4)  Mount the SimpleFS filesystem and set it to be the lib
**   (5)  Do Tk_Init, Tcl_Init, etc
**   (6)  Run main.tcl
**        main.tcl will handle the next steps.
**   (7)  If the first argument is a readable file try to run that script.
**   (8)  Else start a tkCon console.
*/
#include <unistd.h>
#include <stdlib.h>
#include "tcl.h"
#if defined(_WIN32)
#include "tclWinPort.h"
#include "tkWinPort.h"
#endif
#include "tclDecls.h"
#include "tk.h"
#include "obfuskitInit.c"

// A few thing that are defined externally

extern int Vfs_Init(Tcl_Interp*);
extern int Memchan_Init(Tcl_Interp*);
extern int Tk_Init(Tcl_Interp*);
extern int Obfuscateobs_Init(Tcl_Interp*);
unsigned char *simplefs_initScript() ;
unsigned char *obsExpand (unsigned char *, int, int *);
// void *Tcl_SetStartupScript (Tcl_Obj *pathPtr, const char *encodingName);

#if defined(_WIN32)
extern int Dde_Init(Tcl_Interp*);
#endif

/* Used to get the current process ID */
#if !defined(_WIN32)
# include <signal.h>
# include <unistd.h>
# define GETPID getpid
#elif !defined(_WIN32_WCE)
# ifndef SQLITE_AMALGAMATION
#  ifndef WIN32_LEAN_AND_MEAN
#   define WIN32_LEAN_AND_MEAN
#  endif
#  include <windows.h>
# endif
# include <io.h>
# define isatty(h) _isatty(h)
# define GETPID (int)GetCurrentProcessId
#endif

// Invoked from Tk_main, which is called from main()
int Simple_AppInit(Tcl_Interp *interp) {
  int tcl_ret;

  Tcl_SetSystemEncoding(NULL, "utf-8");

  tcl_ret = Tk_Init(interp);
  if (tcl_ret == TCL_ERROR) {
     return(tcl_ret);
  }


#ifdef _WIN32
//  Helps for some debugging
//        tcl_ret = Tk_CreateConsoleWindow(interp);
        if (tcl_ret == TCL_ERROR) {
                return(tcl_ret);
        }
#endif
        return(TCL_OK);
}

// The buck starts here.
int main(int argc, char **argv){
  Tcl_Interp *interp;
  int i;
  unsigned char *zScript = 0;
  char zArgc[32];
  int tcl_ret;
  int rc;

  Tcl_FindExecutable(argv[0]);

  Tcl_SetSystemEncoding(NULL, "utf-8");
  interp = Tcl_CreateInterp();

// printf("CREATED interp) INT: %lx\n", (long int)interp);
  sprintf(zArgc, "%d", argc-1);
  Tcl_SetVar(interp,"argc", zArgc, TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"argv0",argv[0],TCL_GLOBAL_ONLY);
  Tcl_SetVar(interp,"argv", "", TCL_GLOBAL_ONLY);
  for(i=1; i<argc; i++){
    // printf("SETTING: %s\n", argv[i]);
    Tcl_SetVar(interp, "argv", argv[i],
        TCL_GLOBAL_ONLY | TCL_LIST_ELEMENT | TCL_APPEND_VALUE);
  }

  // Test argv/argc setting.
  //Tcl_Eval(interp, "puts ARGC-$argc");
  //Tcl_Eval(interp, "puts ARGV-$argv");

  Vfs_Init(interp);
  Memchan_Init(interp);
#if defined(_WIN32)
  Dde_Init(interp);
  Registry_Init(interp);
#endif

  // Enable the obfuscation entry points
   Obfuscateobs_Init(interp);

  zScript =  simplefs_initScript();
// printf("ZSCRIPT %s\n", zScript); fflush(stdout);

  if (Tcl_EvalEx(interp, (const char*)zScript, -1, TCL_EVAL_GLOBAL)!=TCL_OK ) {
    const char *zInfo = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
    if( zInfo==0 ) zInfo = Tcl_GetStringResult(interp);
    fprintf(stderr,"ERR1 %s: %s\n", *argv, zInfo); fflush(stderr);
    return 1;
  }  
  
  free(zScript);

        tcl_ret = Tcl_Init(interp);
	// printf("AA Tcl_Init Ret: %x\n", tcl_ret);
        if (tcl_ret == TCL_ERROR) {
	  // printf("RETURN FAIL\n"); fflush(stdout);
	  Tcl_Eval(interp, "puts $errorInfo");
          return(tcl_ret);
        }

        Tcl_InitStubs(interp, "8.1", 0);

//	  Tcl_Eval(interp, "puts ERR1-$errorInfo");

//	  Tcl_Eval(interp, \
// "set auto_path [list simpliskit simpliskit/tcl/library simpliskit/tk/library simpliskit/tk simpliskit/tcl]");
//	  Tcl_Eval(interp, "puts PostInitAutoPath-$auto_path");

//Tcl_Eval(interp, "puts GLOB[glob simpliskit/main.tcl]"  );
//Tcl_Eval(interp, "puts GLOBERR-$errorInfo"  );
//Tcl_Eval(interp, "set if [open simpliskit/main.tcl]; set d [read $if]; puts $d");
//Tcl_Eval(interp, "puts GLOBERR2-$errorInfo"  );

Tcl_Eval(interp, "set errorInfo {}"  );
//Tcl_Eval(interp, "puts GLOBERR0-$errorInfo"  );
//Tcl_Eval(interp, "puts TMLIST1-[::tcl::tm::path list]"  );
    rc=Tcl_Eval(interp, "foreach f [::tcl::tm::path list] {::tcl::tm::path remove $f}");
//Tcl_Eval(interp, "puts TMLIST2-[::tcl::tm::path list]"  );
//Tcl_Eval(interp, "puts GLOBERR1-$errorInfo"  );
//Tcl_Eval(interp, "puts GLOBERR2-$errorInfo"  );
//Tcl_Eval(interp, "puts TMLIST3-[::tcl::tm::path list]"  );

    rc=Tcl_Eval(interp, "set result [glob simpliskit/main.tcl]");
    if (rc == TCL_OK) {
        Tcl_Obj *result = Tcl_GetObjResult(interp);
	// printf("OBJRESULT: %s\n", Tcl_GetString(result));
        Tcl_SetStartupScript(result, NULL);
    }

#if defined(_WIN32)
// Kludgy as hell, but I can't get Tcl_<encoding> or
// the Microsoft WideChars converters to work.
  wchar_t *tst[50];
  int j;
  for (i=0; i < argc; i++) {
    tst[i]=calloc(strlen(argv[i])*4+4, 1);
    for (j=0; j<= strlen(argv[i]); j++) {tst[i][j] = argv[i][j];}
  }
  Tk_MainEx(argc, (wchar_t **)tst, Simple_AppInit, interp);
#else
  Tk_MainEx(argc, (char **)argv, Simple_AppInit, interp);
#endif

//printf("BACK FROM Tk_Main\n");

// printf("BACK FROM RUN\n"); fflush(stdout);

  return 0;
}

unsigned char *simplefs_initScript() {
//  printf("INIT: %s\n", obsExpand (crypted, cryptedLen));
  int rtnlen;
  return (obsExpand (crypted, cryptedLen, &rtnlen));
}

