lappend auto_path ../tcllib/tcllib/modules/devtools
catch {source parsePE32A.tcl}

oo::class create simpleFS {

  # FSdir is the directory. A dict :
  # path/element {status offset size}
  # status is "file, dir or deleted"
  # offset is offset from start of file
  # size is size in bytes
  
  # DEBUGGING/TESTING
  method debugPuts {args} {
    variable debug
    if {$debug} {
      puts {*}$args
    }
  }
  # DEBUGGING/TESTING
  method getDir {} {
    variable FSdir
    return $FSdir
  }

################################################################
#   method listing {}--
#    Get the keys for use by VFS when I don't want to put the
#  smarts here
# Arguments
#   NONE
# 
# Globals
#   NONE
# 
# Results
#   NONE
# 
  method listing {} {
    variable FSdir
    return [dict keys $FSdir]
  }

  constructor {} {
    variable debug
    variable opens
    variable State
    
    array set State {
      certificate {}
      certificateLoc 0
      certificateSize 0
      FSDirStart 0
    }
    
    set opens {}
    set debug 0
    variable FSdir
    set FSdir ""
  }
  
  destructor {
    [self] unmount
  }
  
  ################################################################
  # method mount {path}--
  #    Mount a filesystem for access
  # Arguments
  #   path	: Path to file with filesystem on it.
  # 
  # Globals
  #   NONE
  # 
  # Results
  #   In Memory filesystem is loaded.
  # 
  method mount32 {path} {
    variable FSdir
    variable fp
    variable baseSize
    variable accessMode

    my debugPuts "mount path: $path"

    
    if {[catch {open $path a+} fp]} {
      set fp [open $path r]
      set ::errorInfo "NOERR"
      set accessMode r
    } else {
      set accessMode a
    }

    fconfigure $fp -translation binary -encoding binary
    seek $fp -8 end
    my debugPuts "TELL: [tell $fp]"
    set id [read $fp 4]
    my debugPuts "ID: $id"
    if {$id eq "TCLF"} {
      set len [read $fp 4]
      binary scan $len i plen
    my debugPuts "ID: $id PLEN: $plen"
      set newOff [expr {($plen+8)*-1}]
      seek $fp $newOff end
    my debugPuts "TELL2: [tell $fp] - SIZ: [file size $path] FROM $newOff"

      set edir [read $fp $plen ]

      set FSdir [obs_expand $edir]
    my debugPuts "READ DIR: $FSdir"
    } else {
      set baseSize [file size $path]
      set FSdir [dict create "/" [list dir 0 $baseSize]]
    }
  }
  


  method mount {path} {
    variable FSdir
    variable fp
    variable State
    variable baseSize
    variable accessMode

    my debugPuts "mount path: $path"

    
    if {[catch {open $path a+} fp]} {
      set fp [open $path r]
      set ::errorInfo "NOERR"
      set accessMode r
    } else {
      set accessMode a
    }

    fconfigure $fp -translation binary -encoding binary
    if {[catch {parsePE32::accessFile $path $fp}]} {
# puts "NOT MSDOS: $::errorInfo"
      seek $fp -12 end
    } else {
      set sectors [parsePE32::getSectorOffset $path]
# puts "SECTORS: $sectors"
      if {[dict get $sectors Certificate VirtualAddress] == 0} {
        seek $fp -12 end
      } else {
        set off [dict get $sectors Certificate VirtualAddress]
	set State(certificateLoc) [dict get $sectors Certificate VirtualAddress]
	set State(certificateSize) [dict get $sectors Certificate Size]
	seek $fp $off
	set State(certificate) [read $fp $State(certificateSize)]
        seek $fp [expr $off - 256]
        set d [read $fp 256] 
        set p [string first "TCLF" $d]

	seek $fp [expr $off - 256 + $p]
      }
    }
    my debugPuts "TELL: [tell $fp]"
    set id [read $fp 4]
    my debugPuts "ID: $id"
    if {$id eq "TCLF"} {
      set len [read $fp 8]
      binary scan $len i plen
    my debugPuts "ID: $id PLEN: $plen"
      set curLoc [tell $fp]

      set newOff [expr {($curLoc - $plen-12)}]
      set State(FSDirStart) $newOff
      seek $fp $newOff

    my debugPuts "TELL2: [tell $fp] - SIZ: [file size $path] FROM $newOff"

      set edir [read $fp $plen ]

      set FSdir [obs_expand $edir]
    my debugPuts "READ DIR: $FSdir"
    } else {
      set baseSize [file size $path]
      set FSdir [dict create "/" [list dir 0 $baseSize]]
    }
  }
  
  method access {path mode} {
    variable accessMode
    variable FSdir
 my debugPuts "FS_ACCESS: PATH $path -- MODE $mode == [vfs::accessMode $mode] : KEYS: '[dict keys $FSdir $path]'"

# my pprint
# puts "KEYS: '[dict keys $FSdir ]'"
#  puts "FS_ACCESS: $path -- $mode : FOLDERKEYS: '[dict keys $FSdir $path/*]'"
# puts "FSdir: $FSdir"
    set rtn 0
    if {[llength [dict keys $FSdir $path]] != 0} {
      # Have file
      set rtn 1
    }

    if {([llength [dict keys $FSdir $path/*]] != 0) && \
        (($mode == 0) || ($mode == 4))} {
      # Have folder
      set rtn 1
    }

    my debugPuts "access: RETURN $rtn"
    return $rtn
  }

  method unmount {} {
    variable fp
    my debugPuts "unmount"
    close $fp
  }

  method writeDir {} {
    variable FSdir
    variable fp
    variable State
    my debugPuts "writeDir"

    set lastKey [lindex [dict keys $FSdir] end]
    my debugPuts "LAST: $lastKey VALS: [dict get $FSdir $lastKey]"
    lassign [dict get $FSdir $lastKey] type offset size
    seek $fp [expr {$offset+$size}] start
# puts "WRITING DIR AT: [format %xx [expr {$offset+$size}]] ([expr {$offset+$size}])"
    set edir [obs_shrink $FSdir]
    set len [string length $edir]
    puts -nonewline $fp $edir
# puts "TCLF AT [format %xx [tell $fp]] ([tell $fp])"
    puts -nonewline $fp "TCLF[binary format w $len]"
    
    if {$State(certificateLoc) != 0} {
# puts "LOC: $State(certificateLoc) - [format %xx $State(certificateLoc)]"
      set pos [tell $fp]
      set pad [expr (64-$pos%64)]
# puts "WRITING PAD AT: [format %xx $pos] ($pos) PADLEN: $pad"
      seek $fp $pos 
      puts -nonewline $fp [string repeat [binary format c 0] $pad]
      set certPos [tell $fp]
# puts "WRITING CERT AT: [format %xx $certPos] ($certPos)"
      puts -nonewline $fp $State(certificate);
      parsePE32::updateDataDir Certificate $certPos $State(certificateSize) $fp
    }
    flush $fp
  }
  
  method open {name op} {

    variable opens
    variable FSdir
    my debugPuts " Open: name: $name OP: $op"

    if {![dict exists $FSdir $name]} {
      my debugPuts "Open: FAIL $name NOT EXIST"
      # eexist is 17 per errno.h
      # enent is 2 per errno.h - no such file or dir
# puts "RETURN CODE 2"
      return -code 2 "$name NOT EXIST"
    }

catch {info level -1} frm
# puts "OPEN FRM: $frm"


    if {[dict exists $opens $name]} {
      error "Only one open per file"
    }
    if {$op eq "r"} {
      lassign [dict get $FSdir $name] type offset length
      set start $offset
      set end [expr {$start + $length}]
    } else {
      set start 0
      set end 0
    }
    set id fs[expr {int(10000*rand())}]
    dict append opens $id [list $name $op $start $end]
# puts "OPENS: $opens"
# puts "ID: $id"; flush stdout
    set ::errorInfo "MT1"
    my debugPuts " Open: RTN $id"
    return -code 0 $id
  }
  
  method close {id} {
    variable opens
    my debugPuts " CLOSE: id $id"
    set opens [dict remove $opens $id]
  }
  
  method puts {id data} {
    variable opens
    my debugPuts " PUTS: $id $data"

    if {![dict exists $opens $id]} {
      return -code 1 "Not open"
    }
    lassign [dict get $opens $id] name op start end
    if {$start != 0} {
      return -code 1 "Invalid file"
    }
    my writeFile $name $data
    my close $id
  }

  method gets {id} {
    variable opens
    variable fp
    my debugPuts "gets: ID $id"
    lassign [dict get $opens $id] name op start end
    if {$start >= $end} {
      return -code 1 "EOF"
    }
    seek $fp $start start
    set rtn [gets $fp]

    incr start [string length $rtn]
    incr start
    dict set opens $id [list $name $op $start $end]
    my debugPuts "gets: RTN $rtn"
    return $rtn
  }
  
  method addToDir {name newType newSize} {
    variable FSdir
    my debugPuts "A1 DIR: $FSdir"
# puts "FSDIR: $FSdir"
    set lastKey [lindex [dict keys $FSdir] end]
    my debugPuts "LAST: $lastKey VALS: [dict get $FSdir $lastKey]"
    lassign [dict get $FSdir $lastKey] type offset size
    if {($lastKey eq "FREE") && ($size >= $newSize)} {
      set newStart $offset
      set newFreeVals [list deleted [expr {$offset + $newSize}] [expr {$size - $newSize}]]
      set FSdir [dict remove $FSdir FREE]
    } else {
      set newStart [expr {$offset+$size}]
      set newFreeVals ""
    }

    if {[dict exists $FSdir $name]} {
      my delFile $name
    }
    dict append FSdir $name [list $newType $newStart $newSize]
    if {$newFreeVals ne ""} {
      dict append FSdir FREE $newFreeVals
    }
    my debugPuts "ADD DIR: $FSdir"
  }
  
  method findInDir {name} {
    variable FSdir
    error "findInDir {name} NOT IMPLEMENTED"
  }

  method writeFile {name data {preshrunk 0} } {
    variable FSdir
    variable fp
    my debugPuts "writeFile: $name LEN [string length $data]"
    if {!$preshrunk} {
      set data [obs_shrink $data]
    }
    
    if {[string first "/" $name] != 0} {
      set name "/$name"
    }

    my addToDir $name file [string length $data]
    lassign [dict get $FSdir $name] type offset len
    seek $fp $offset start
    puts $fp $data
    my writeDir
  }

  method readFile {name} {
    variable FSdir
    variable fp
    my debugPuts "readFile: $name"

    if {![dict exists $FSdir $name]} {
      my debugPuts "readFile: FAIL $name NOT EXIST"
      # eexist is 17 per errno.h
      # 2 is enoent, used to be enofile

      return -code 2 "$name NOT EXIST"
    }
    lassign [dict get $FSdir $name] type offset len
    seek $fp $offset start
    set d [read $fp $len]
    set d [obs_expand $d]
    my debugPuts "readFile: RTN LEN: [string length $d]"

    return $d
  }
  
  method pprint {} {
    variable FSdir
    dict for {k v} $FSdir {
      lassign $v type off size
      puts [format "%-25s %8s %8d %8d" $k  $type $off $size]
    }
  }
  
  method doDir {{pattern {}} {type 0}} {
    variable FSdir
    set rtn {}
    if {$type == 4} {
      set pattern $pattern/*
    }
# puts "DIR: PATTERN --$pattern-- TYPE: $type "
# puts "DIR FROM: [info level -1]"

   if {$pattern eq ""} {
      set rtn [dict keys $FSdir]
    } elseif {([string first * $pattern] < 0)} {
      # Exact file name search
      set rtn [lsearch -inline [dict keys $FSdir] $pattern]
# puts "RTN1: $rtn"
    } 
    
    if {($rtn eq "") && ([string first * $pattern] < 0) && ($type == 4)} {
      # Failed first test
      # Might be a directory search with no wildcard
      set rtn [lsearch -inline $FSdir $pattern/*]
# puts "RTN2: $rtn"
      if {$rtn ne ""} {
        set rtn $pattern
      }
    }
catch {
    if {$rtn eq ""} {
        set regex ^[string map {* [^/\]+} $pattern]
      my debugPuts "PATTERN: $pattern REGEX: $regex KEYS: [dict keys $FSdir]"
# puts "PATTERN: $pattern REGEX: $regex MATCHKEYS: [lsearch -all -inline -regexp [dict keys $FSdir] $regex ]"
      
      set match [lsearch -all -inline -regexp  [dict keys $FSdir] $regex ]
# puts "FS DIR MATCH: $match FROM $regex"
      foreach r $match {
        if {$type == 4} {
# puts "SET M1: [file dirname $r] FROM $r"
#	  set m1 [file dirname $r]
          set regex [string map {* [^/\]+} [string range $pattern 0 end-2]]
          regexp $regex $r m1 m2
	  set tp directory
	} else {
          if {[string first "DEL" $r] == 0} {continue}
          regexp $regex $r m1 m2
# puts "M1: $m1 M2: $m2"
          if {[dict exists $FSdir $m1]} {
	    lassign [dict get $FSdir $m1] tp off siz
# puts "TYPE: $tp"
	  } else {
	    set tp ""
	  }
	}
	if {($tp ne "deleted") && ([lsearch $rtn $m1] < 0)} {
	  lappend rtn $m1
	}
      }

      if {[set p [lsearch $rtn "/"]] >= 0} {
        incr p
	set rtn [lrange $rtn $p end]
      }
    }
} foo
# puts "FOO: $foo"
# puts "DIR: return '$rtn'"
    my debugPuts "dir RETURN $rtn"
      return $rtn
  }
    
  method stat {name} {
    variable FSdir
    my debugPuts "stat name: $name"
    # puts "METHOD stat: NAME: $name FROM [info level -1]"
    set name [string map {// /} $name]

    set file [my doDir $name]
    set children [my doDir $name/*]
    
    if {[llength $children] > 0} {
      set type directory
      set size [llength $children]
    } else {
      set type file
      if {![dict exists $FSdir $name]} {
        error "No File $name"
      }
      
      lassign [dict get $FSdir $name] type off size
      if {$type eq "dir"} {set type "directory"}
    }

    my debugPuts "STAT    return [dict create type $type size $size]"
    return [dict create type $type size $size]
  }

  method delFile {{pattern {}}} {
    variable FSdir
    my debugPuts "delFile pattern $pattern" 
    if {$pattern eq ""} {
      set keys [dict keys $FSdir]
    } else {
      set keys [dict keys $FSdir $pattern]
    }
#  puts "delFile $pattern -> $keys"
    foreach k $keys {
      lassign [dict get $FSdir $k] type off size
# puts "T: $type O: $off S: $size"
      dict set FSdir $k [list "deleted" $off $size]
# puts "MOD: [dict get $FSdir $k]"
      lset FSdir [lsearch $FSdir $k] DEL_$k
    }
    # my pprint
  }
  
  method mergeDEL {} {
    variable FSdir
    my debugPuts "mergeDEL"

    dict for {k v} $FSdir {
      lassign $v tmpType tmpOff tmpSize
      if {$tmpType eq "deleted"} {
        lassign $v newType newOff newSize
        if {$off > 0} {
	  set size [expr $size + $newSize]
	  dict set FSdir $name [list $type $off $size]
	  set FSdir [dict remove $FSdir $k]
	} else {
          lassign $v type off size
	  set name $k
	}
      } else {
        lassign {-1 -1 -1} type off size
      }
    }
  }
  
  method compressFiles {} {
    variable FSdir
    variable fp
    my debugPuts "compressFiles"
    
    my mergeDEL
    set freeSpace 0
    dict for {k v} $FSdir {
      lassign $v thisType thisOff thisSize
      if {$thisType eq "deleted"} {
        lassign $v newType newOff newSize
	set keys [dict keys $FSdir]
	set p [lsearch $keys $k]
	incr p
	foreach k2 [lrange $keys $p end] {
	  lassign [dict get $FSdir $k2] type off size
	  seek $fp $off start
	  set d [read $fp $size]
	  seek $fp $newOff start
	  puts -nonewline $fp $d
	  dict set FSdir $k2 [list $type $newOff $size]
	  incr newOff $size
	}
	set FSdir [dict remove $FSdir $k]
	dict append FSdir $k [list deleted $newOff $newSize]
      }
    }

    my mergeDEL

    set last [lindex [dict keys $FSdir] end]
    set v [dict get $FSdir $last]
    
    lassign $v stat off size
    if {$stat eq "deleted"} {
      set FSdir [dict remove $FSdir $last]
      dict append FSdir FREE $v
    }
  }
  
  method fileOrDir {path} {
    variable FSdir
    my debugPuts "fileOrDir $path"
    
    set rtn ""
    if {[dict exists $FSdir $path]} {
      set rtn file
    }
    if {[llength [dict keys $FSdir $path/*]] > 0} {
      set rtn dir
    }

    my debugPuts "fileOrDir RTN $rtn"

    if {$rtn eq ""} {
      error "$path does not exist in [dict keys $FSdir]"
    } else {
      return $rtn
    }
  }
}

if {0} {
set of [open xx w]
puts $of "test of nothing"
close $of
set fs [simpleFS create newFS]
$fs mount xx
$fs writeFile newFile "THIS IS ANOTHER TEST"

$fs pprint
$fs writeDir
$fs unmount


$fs mount xx
 puts "READ: [$fs readFile /newFile]"

$fs unmount

$fs mount xx
for {set i 0} {$i < 5} {incr i} {
  $fs writeFile testing_$i "THIS IS TEST $i---"
}
puts "NEW DIR: [$fs dir]"
puts "NEW Test DIR: [$fs dir /test*]"

foreach d [$fs dir /test*] {
  puts "READIN: $d: [$fs readFile $d]"
}

$fs delFile /testing_2
puts "Post Del DIR: [$fs dir ]"
puts "POST DEL:"
$fs pprint
$fs delFile /testing_1
puts ""
puts "POST DEL2:"
$fs pprint

$fs mergeDEL

puts "\nPOST MERGE:"
$fs pprint

flush stdout
$fs compressFiles
flush stdout

puts "\nPOST COMPDIR: $$fs dir"
flush stdout

puts "\nPOST COMPR:"
$fs pprint

puts "READING 3: /testing_3: [$fs readFile /testing_3]"

$fs writeFile testing_5 "THIS IS TEST 5-"

puts "\nPOST New Write:"
$fs pprint

puts "READING 5: /testing_5: [$fs readFile /testing_5]"

}
